import java.awt.Button;
import java.awt.Frame;
import java.awt.TextField;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.util.Timer;
import java.util.TimerTask;

public class Tris extends Frame implements ActionListener, WindowListener, MouseListener {
	private static final long serialVersionUID = -4231045267611273745L;
	
	//Net
	private static final int DEFAULT_PORT = 6000;
	private DatagramSocket sock;
	private static final int BUFFER_SIZE = 256;
	private static final int TIMEOUT = 100;
	private static final int REFRESH = 500;
	private boolean changing = false;
	
	//Interface
	private Cella[] grid       = new Cella[9];
	private TextField lNick    = new TextField("Giocatore1");
	private TextField lAddr    = new TextField();
	private TextField lPort    = new TextField();
	private Button    lBtn     = new Button("OK");
	private TextField rNick    = new TextField("Disconnesso");
	private TextField rAddr    = new TextField();
	private TextField rPort    = new TextField();
	private Button    rBtn     = new Button("OK");
	private Button    btnStart = new Button("Inizia partita");
	private Button    btnEnd   = new Button("Smetti di giocare");
	private static final int SIZE = 100;
	private static final int OFFX = 20;
	private static final int OFFY = 40;
	
	//Game
	private String myName = "Giocatore1";
	private InetSocketAddress player;
	private boolean playing = false;
	private boolean requesting = false;
	private boolean wait = false;
	private boolean finished = true;
	private int mySymbol;
	private int otherSymbol;
	private int count = 0;
	
	//Protocol
	private static final byte   REQUEST  = 1;
	private static final byte   PUT      = 2;
	private static final byte   NEW      = 3;
	private static final byte   ACCEPT   = 4;
	private static final byte   REFUSE   = 5;
	private static final String ENCODING = "US-ASCII";
	
	private Tris() {
		this.setSize(440,500);
		this.setLayout(null);
		int x = OFFX;
		int y = OFFY;
		lNick.setBounds(x, y, 100, 20);
		this.add(lNick);
		x += 110;
		lAddr.setBounds(x, y, 200, 20);
		lAddr.setEditable(false);
		this.add(lAddr);
		x += 210;
		lPort.setBounds(x, y, 50, 20);
		this.add(lPort);
		x += 60;
		lBtn.setBounds(x, y, 20, 20);
		lBtn.addActionListener(this);
		this.add(lBtn);
		x = OFFX;
		y += 20;
		rNick.setBounds(x, y, 100, 20);
		rNick.setEnabled(false);
		this.add(rNick);
		x += 110;
		rAddr.setBounds(x, y, 200, 20);
		this.add(rAddr);
		x += 210;
		rPort.setBounds(x, y, 50, 20);
		this.add(rPort);
		x += 60;
		rBtn.setBounds(x, y, 20, 20);
		rBtn.addActionListener(this);
		this.add(rBtn);
		x = OFFX;
		y += 30;
		btnStart.setBounds(x, y, 195, 20);
		btnStart.addActionListener(this);
		btnStart.setEnabled(false);
		this.add(btnStart);
		x += 205;
		btnEnd.setBounds(x, y, 195, 20);
		btnEnd.addActionListener(this);
		btnEnd.setEnabled(false);
		this.add(btnEnd);
		x = OFFX;
		y += 30;
		Cella.init();
		for (int i = 0; i < grid.length; i++) {
			grid[i] = new Cella();
			grid[i].setBounds(x+(i%3)*SIZE, y+(i/3)*SIZE, SIZE, SIZE);
			grid[i].addMouseListener(this);
			this.add(grid[i]);
		}
		try {
			sock = new DatagramSocket(DEFAULT_PORT);
			sock.setSoTimeout(TIMEOUT);
			lAddr.setText(InetAddress.getLocalHost().toString());
			lPort.setText(""+DEFAULT_PORT);
			
			Timer timer = new Timer();
			timer.schedule(new TimerTask() {
				@Override
				public void run() {
					if (changing) return;
					byte[] buf = new byte[BUFFER_SIZE];	
					DatagramPacket packet = new DatagramPacket(buf, BUFFER_SIZE);
					try {
						sock.receive(packet);
						usePacket(packet);
					} catch(SocketTimeoutException e) {
					} catch (IOException e) {
						e.printStackTrace();
					} 
				}
			}, 0, REFRESH);
		} catch (SocketException e) {
			e.printStackTrace();
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
		this.addWindowListener(this);
		this.setVisible(true);
	}
	
	public static void main(String[] args) {
		new Tris();
	}
	
	private void usePacket(DatagramPacket packet) {
		byte[] buffer = packet.getData();
		InetSocketAddress remoteAddr = (InetSocketAddress) packet.getSocketAddress();
		if (buffer.length < 1) return;
		
		String name;
		
		
		/*try {
			name = new String(buffer, ENCODING);
			name = name.substring(1,name.indexOf(0));
			System.out.println("OP"+buffer[0]+": "+name);
		} catch (UnsupportedEncodingException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}*/
		
		switch(buffer[0]) {
			case REQUEST:
				if (playing) {
					refuse(remoteAddr);
					return;
				}
				name = "";
				try {
					name = new String(buffer, ENCODING);
					name = name.substring(1,name.indexOf(0));
				} catch (UnsupportedEncodingException e) {
					e.printStackTrace();
				}
				Richiesta req = new Richiesta(name, remoteAddr.getAddress());
				if (req.getResponse()) {
					start(name, remoteAddr);
					accept(remoteAddr);
				}
				else {
					refuse(remoteAddr);
				}
				return;
			case PUT:
				if (playing && remoteAddr.equals(player) && wait && buffer.length > 1) {
					if (putSymbol(false, buffer[1]))
						wait = false;
				}
				return;
			case NEW:
				if (finished && playing && remoteAddr.equals(player)) {
					startNew(true);
				}
				return;
			case ACCEPT:
				if (!requesting || !remoteAddr.equals(player)) return;
				name = "";
				try {
					name = new String(buffer, ENCODING);
					name = name.substring(1,name.indexOf(0));
				} catch (UnsupportedEncodingException e) {
					e.printStackTrace();
				}
				new Risposta(name, remoteAddr.getAddress(), true);
				requesting = false;
				playing = true;
				start(name, remoteAddr);
				
				
				startNew(true);
				return;
			case REFUSE:
				if (!remoteAddr.equals(player)) return;
				new Risposta("", remoteAddr.getAddress(), false);
				playing = false;
				if (!requesting) stop();
				return;
			default:
				return;
		}
	}
	
	private void start(String name, InetSocketAddress addr) {
		finished = true;
		wait = false;
		playing = true;
		lNick.setEnabled(false);
		lAddr.setEnabled(false);
		lPort.setEnabled(false);
		lBtn.setEnabled(false);
		rNick.setText(name);
		rAddr.setText(addr.getAddress().toString());
		rPort.setText(addr.getPort()+"");
		rAddr.setEnabled(false);
		rPort.setEnabled(false);
		rBtn.setEnabled(false);
		btnStart.setEnabled(true);
		btnEnd.setEnabled(true);
	}
	
	private void stop() {
		finished = true;
		wait = false;
		playing = false;
		lNick.setEnabled(true);
		lAddr.setEnabled(true);
		lPort.setEnabled(true);
		lBtn.setEnabled(true);
		rNick.setText("Disconnesso");
		rAddr.setText("");
		rPort.setText("");
		rAddr.setEnabled(true);
		rPort.setEnabled(true);
		rBtn.setEnabled(true);
		btnStart.setEnabled(false);
		btnEnd.setEnabled(false);
	}
	
	private void sendRequest() {
		player = new InetSocketAddress(rAddr.getText(),Integer.parseInt(rPort.getText()));
		byte[] buffer1;
		try {
			buffer1 = myName.getBytes(ENCODING);
			int len = buffer1.length+2>BUFFER_SIZE?BUFFER_SIZE:buffer1.length+2;
			byte[] buffer2 = new byte[len];
			buffer2[0] = REQUEST;
			for (int i = 0; i < len-2; i++) {
				buffer2[i+1] = buffer1[i];
			}
			buffer2[len-1] = 0;
			DatagramPacket packet;
			packet = new DatagramPacket(buffer2, len, player);
			sock.send(packet);
			requesting = true;
		} catch (SocketException e) {
			e.printStackTrace();
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private void accept(InetSocketAddress addr) {
		try {
			byte[] buffer1 = myName.getBytes(ENCODING);
			int len = buffer1.length+2>BUFFER_SIZE?BUFFER_SIZE:buffer1.length+2;
			byte[] buffer2 = new byte[len];
			buffer2[0] = ACCEPT;
			for (int i = 0; i < len-2; i++) {
				buffer2[i+1] = buffer1[i];
			}
			buffer2[len-1] = 0;
			DatagramPacket packet;
			packet = new DatagramPacket(buffer2, len, addr);
			sock.send(packet);
			player = addr;
			
			startNew(false);
		} catch (SocketException e) {
			e.printStackTrace();
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private void refuse(InetSocketAddress addr) {
		try {
			byte[] buffer = new byte[] {REFUSE};
			DatagramPacket packet = new DatagramPacket(buffer, 1, addr);
			sock.send(packet);
		} catch (SocketException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private void sendNew() {
		try {
			byte[] buffer = new byte[] {NEW};
			DatagramPacket packet = new DatagramPacket(buffer, 1, player);
			sock.send(packet);
			startNew(false);
		} catch (SocketException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private void startNew(boolean first) {
		count = 0;
		btnStart.setEnabled(false);
		for (Cella c : grid) c.reset();
		finished = false;
		if (first) {
			wait=false;
			mySymbol = Cella.X;
			otherSymbol = Cella.O;
		} else {
			wait=true;
			mySymbol = Cella.O;
			otherSymbol = Cella.X;
		}
	}
	
	private boolean putSymbol(boolean mine, int index) {
		if (index < 0 || index > 8) return false;
		
		count++;
		int symbol = otherSymbol;
		if (mine)
			symbol = mySymbol;
		
		if (mine && grid[index].setStatus(mySymbol)) {
			wait = true;
		} else if (grid[index].setStatus(otherSymbol)) {
			wait = false;
		} else {
			count--;
			return false;
		}
		
		int x = index%3;
		int y = index/3;
		
		boolean end = true;
		for (int i = 0; i < 3; i++) {
			if (grid[3*i+x].getStatus() != symbol) {
				end = false;
				break;
			}
		}
		if (end) {
			new Fine(mine?Fine.WIN:Fine.LOSE);
			btnStart.setEnabled(true);
			finished = true;
			return true;
		}
		end = true;
		for (int i = 0; i < 3; i++) {
			if (grid[3*y+i].getStatus() != symbol) {
				end = false;
				break;
			}
		}
		if (end) {
			new Fine(mine?Fine.WIN:Fine.LOSE);
			btnStart.setEnabled(true);
			finished = true;
			return true;
		}
		if (x == y) {
			end = true;
			for (int i = 0; i < 3; i++) {
				if (grid[4*i].getStatus() != symbol) {
					end = false;
					break;
				}
			}
			if (end) {
				new Fine(mine?Fine.WIN:Fine.LOSE);
				btnStart.setEnabled(true);
				finished = true;
				return true;
			}
		}
		if (x + y == 2) {
			end = true;
			for (int i = 1; i <= 3; i++) {
				if (grid[2*i].getStatus() != symbol) {
					end = false;
					break;
				}
			}
			if (end) {
				new Fine(mine?Fine.WIN:Fine.LOSE);
				btnStart.setEnabled(true);
				finished = true;
				return true;
			}
		}
		if (count == 9) {
			new Fine(Fine.DRAW);
			btnStart.setEnabled(true);
			finished = true;
		}
		return true;
	}
	
	@Override
	public void actionPerformed(ActionEvent arg0) {		
		if (arg0.getSource() == lBtn) {
			try {
				changing = true;
				if (sock.isBound()) sock.close();
				sock = new DatagramSocket(Integer.parseInt(lPort.getText()));
				sock.setSoTimeout(TIMEOUT);
				myName = lNick.getText();
				changing = false;
			} catch (NumberFormatException e) {
				e.printStackTrace();
			} catch (SocketException e) {
				e.printStackTrace();
			}
		} else if (arg0.getSource() == rBtn) {
			if (requesting) {
				refuse(player);
			}
			sendRequest();
		} else if (arg0.getSource() == btnStart) {
			sendNew();
		} else if (arg0.getSource() == btnEnd) {
			stop();
			refuse(player);
		}
	}

	@Override
	public void windowActivated(WindowEvent arg0) {		
	}

	@Override
	public void windowClosed(WindowEvent arg0) {
		System.exit(0);
	}

	@Override
	public void windowClosing(WindowEvent arg0) {
		this.dispose();
	}

	@Override
	public void windowDeactivated(WindowEvent arg0) {
	}

	@Override
	public void windowDeiconified(WindowEvent arg0) {
	}

	@Override
	public void windowIconified(WindowEvent arg0) {
	}

	@Override
	public void windowOpened(WindowEvent arg0) {		
	}

	@Override
	public void mouseClicked(MouseEvent arg0) {
	    if (!finished && !wait) {
			for (int i = 0; i < grid.length; i++) {
				if (arg0.getSource() == grid[i] && grid[i].getStatus() == Cella.VOID) {
					byte[] buffer = new byte[] {PUT, (byte) i};
					DatagramPacket packet;
					try {
						packet = new DatagramPacket(buffer, 2, player);
						sock.send(packet);
					} catch (SocketException e) {
						e.printStackTrace();
					} catch (IOException e) {
						e.printStackTrace();
					}
					
					putSymbol(true,i);
				}
			}
	    }
	}

	@Override
	public void mouseEntered(MouseEvent arg0) {
	}

	@Override
	public void mouseExited(MouseEvent arg0) {
	}

	@Override
	public void mousePressed(MouseEvent arg0) {
	}

	@Override
	public void mouseReleased(MouseEvent arg0) {
	}
}
